(* Copyright 2008 Martin Korp, Christian Sternagel, Harald Zankl
 * GNU Lesser General Public License
 *
 * This file is part of TTT2.
 * 
 * TTT2 is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * TTT2 is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with TTT2. If not, see <http://www.gnu.org/licenses/>.
 *)

(*** MODULES ******************************************************************)
module F = Format;;
module M = Monad;;
module O = Util.Option;;
module Pos = Position;;

(*** OPENS ********************************************************************)
open Prelude;;
open Util;;

(*** MODULES ******************************************************************)
module Make (L : LABEL) = struct
 (*** MODULES *****************************************************************)
 module E = Elogic.Make (L);;
 module M = M.Make (L);;
 (* module Parser = Parser.Make (L);; *)
 (* module Xml = Parser.Xml;; *)
 module S = Signature.Make (L);;
 module Sub = Substitution.Make (L);;
 module Term = Term.Make (L);;
 module Rule = Rule.Make (L);;
 module Trs = Trs.Make (L);;
 module Rewrite = Rewrite.Make (L);;

 (*** TYPES *******************************************************************)
 type 'a m = 'a M.t
 type term = Term.t = Var of Variable.t | Fun of Function.t * term list;;
 type rule = Rule.t;;
 type step = Rewrite.step;;
 type trs = Trs.t;;
 type t = (step * step) * (step list * step list) list;;
 
 (*** FUNCTIONS ***************************************************************)
 (* Parsers *)
 let (<?>) = Util.(<?>);;
 let (>>=) = M.(>>=);;
 let (>>) = M.(>>);;
 
 (* Iterators *)
 (* Scan Functions *)
 (* Constructors and Destructors *)
 let peak (p,_) = p;;
 let joins (_,js) = js;;
 let make p js = (p,js);;

 (* Compare Functions *)
 let compare = compare;;
 let equal r r' = compare r r' = 0;;

 let critical_pair d = Pair.map Rewrite.step_get_reduct (peak d);;
  
 let diagram ?(m=0) trs ((r1,p,r2) as o) =
  Rewrite.critical_peak o >>= fun (t,s,u) ->
  (* let t = Rule.rewrite s p r1 in *)
  (* let u = Rule.rewrite s Position.root r2 in *)
  let l = Rewrite.step_make trs s p r1 in
  let r = Rewrite.step_make trs s Position.root r2 in
  if m < 0 then (*do not try to join*)
   M.return (Some ((l,r), [])) 
  else
   try
   let js = Option.the (Rewrite.min_joins ~m:m t u trs) in
   let ss = List.map (Pair.map (Rewrite.sequence trs)) js in
   M.return (Some ((l,r), ss))
   with
    | Failure "no content" -> M.return None
 ;;

 let critical_diagrams ?(m=0) trs = 
  Trs.overlaps trs >>= fun os ->
  M.map (diagram ~m:m trs) os >>= fun ds ->
  try
   M.return (Some (List.map Option.the ds))
  with
   | Failure "no content" -> M.return None
 ;;
 
 let join_map f j = Pair.map (List.map f) j;; 
 (*apply function f to all steps in [d]*)
 let map f d = 
  let p = Pair.map f (peak d) in
  let js = List.map (join_map f) (joins d) in
  make p js
 ;;
 
 (* label a diagram *)
 let rec mapl_seq f n = function
  | [] -> []
  | s::ss -> 
    let m,s = f n s in
    let ss = mapl_seq f m ss in
    (s::ss)
 ;;
 
 let mapl_j f nl nr (ls,rs) = 
  let ls = mapl_seq f nl ls in
  let rs = mapl_seq f nr rs in
  (ls,rs)
 ;;
  
 let mapl f n d = 
  let (l,r) = peak d in
  let nl,l = f n l in
  let nr,r = f n r in
  let js = List.map (mapl_j f nl nr) (joins d) in
  (make (l,r) js)
 ;;

 let rec mapm_seq f n = function
  | [] -> M.return []
  | s::ss -> 
   f n s >>= fun (m,s) -> 
   mapm_seq f m ss >>= fun ss ->
   M.return (s::ss)
 ;;
 
 let mapm_j f nl nr (ls,rs) = 
  mapm_seq f nl ls >>= fun ls ->
  mapm_seq f nr rs >>= fun rs ->
  M.return (ls,rs)
 ;;
  
 let mapm f n d = 
  let (l,r) = peak d in
  f n l >>= fun (nl,l) ->
  f n r >>= fun (nr,r) ->
  M.map (mapm_j f nl nr) (joins d) >>= fun js ->
  M.return (make (l,r) js)
 ;;

 let label label_step n d = mapl label_step n d;;
 let labelm label_stepm n d = mapm label_stepm n d;;
 let map f d =  mapl (fun n s -> (~-1,f s)) (~-1) d;;

 (*test decreasingness*)

 let lex_eq ls rs = List.for_all2 (=) ls rs;;

 let rec lex_gt ls rs = match ls,rs with
  | l::ls,r::rs when l > r -> true
  | l::ls,r::rs when l = r -> lex_gt ls rs
  | _ -> false
 ;;

 let is_decreasing_l_strict alpha beta ls = 
  List.for_all (fun l -> lex_gt alpha l || lex_gt beta l) ls
 ;;

 let rec is_decreasing_l alpha beta = function
  | [] -> true
  | l::ls when lex_gt alpha l -> is_decreasing_l alpha beta ls
  | l::ls when lex_eq beta l -> is_decreasing_l_strict alpha beta ls
  | ls -> is_decreasing_l_strict alpha beta ls
 ;;

 let is_decreasing_s dec alpha beta ss = 
  dec alpha beta (List.map Rewrite.step_get_lab ss)
 ;;

 let is_decreasing_j dec alpha beta (ls,rs) =
  is_decreasing_s dec alpha beta ls &
  is_decreasing_s dec beta alpha rs
 ;;

 let is_decreasing_aux dec d = 
  let (l,r) = peak d in
  let (nl,nr) = (Rewrite.step_get_lab l, Rewrite.step_get_lab r) in
  List.exists (is_decreasing_j dec nl nr) (joins d)
  ;;

 let is_decreasing d = is_decreasing_aux is_decreasing_l d;;
 let is_decreasing_strict d = is_decreasing_aux is_decreasing_l_strict d;;

 let is_inner_parallel_closed trs s t = List.mem t (Rewrite.parallel trs s);;
 let is_outer_parallel_closed trs s t = 
  List.intersect (Rewrite.parallel trs s) (Rewrite.parallel trs t) <> []
 ;;

 let is_parallel_closed trs d = 
 let (l,r) = peak d in
 let (s,t) = Pair.map Rewrite.step_get_reduct (l,r) in
 if Position.is_root (Rewrite.step_get_pos l) then
  is_outer_parallel_closed trs s t
 else
  is_inner_parallel_closed trs s t

 ;;

 let one2many d = List.map (fun join -> make (peak d) [join]) (joins d);;
 let many2one ds = make (peak (List.hd ds)) (List.flat_map joins ds);;
 
 let filter p d = many2one (List.filter p (one2many d));;

 let compare_short d1 d2 =
  let [(ls1,rs1)] = joins d1 in
  let [(ls2,rs2)] = joins d2 in
  compare (List.length (ls1@rs1)) (List.length (ls2@rs2))
 ;;

 let find p d = List.hd (List.sort compare_short (one2many (filter p d)));;
 
 (*
 (* Printers *)
 let fprintf fmt r =
  (* F.fprintf fmt "@[%a@ ->@ %a@]" Term.fprintf (lhs r) Term.fprintf
  (rhs r) *)
  ();
 ;;
*)

 let fprintfm_peak fmt d = 
  let (l,r) = peak d in
  Term.fprintfm fmt (Rewrite.step_get_reduct l) >>= fun _ ->
  F.fprintf fmt " <-%d|%a[%a]- " 
   (Rewrite.step_get_rule_idx l)
   Position.fprintf (Rewrite.step_get_pos l)
   Rewrite.fprintf_lab (Rewrite.step_get_lab l);
  Rewrite.fprintfm_step ~reduct:true fmt r >>= fun _ ->
  M.return ();
 ;;

 let fprintfm_join fmt (ls,rs) = 
  Rewrite.fprintfm_sequence fmt ls >>= fun _ ->
  Format.fprintf fmt "@\n";
  Rewrite.fprintfm_sequence fmt rs >>= fun _ ->
  M.return ();
 ;;

 let fprintfm_joins fmt d = 
  M.iteri (fun i j -> 
   if i > 0 then Format.fprintf fmt "@\n---@\n";
   fprintfm_join fmt j >>= fun _ -> M.return ()) (joins d)
  ;;

 let fprintfm fmt d =
  F.fprintf fmt "@[@[<1>peak:@\n";
  fprintfm_peak fmt d >>= fun _ ->
  F.fprintf fmt "@]@\n";
  F.fprintf fmt "@[<1>joins:@\n";
  fprintfm_joins fmt d >>= fun _ ->
  F.fprintf fmt "@]@]";
  M.return ();
 ;;

(*
 let to_string = F.flush_str_formatter <.> fprintf F.str_formatter;;

 let to_stringm r =
  (* fprintfm F.str_formatter r >>= (M.return <.> F.flush_str_formatter) *)
  M.return ();
 ;; 
 *)
end
